# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Facilities for managing and reporting errors.
"""
import os
import sys

import dbus

from dbus_client_gen import DbusClientMissingInterfaceError
from dbus_client_gen import DbusClientMissingPropertyError
from dbus_client_gen import DbusClientMissingSearchPropertiesError
from dbus_client_gen import DbusClientUniqueResultError
from dbus_client_gen import DbusClientUnknownSearchPropertiesError

from ._actions import interface_name_to_common_name

_DBUS_INTERFACE_MSG = (
    "The version of stratis you are running expects a different "
    "D-Bus interface than the one stratisd provides. Most likely "
    "you are running a version that requires a newer version of "
    "stratisd than you are running.")

_STRATIS_CLI_BUG_MSG = ("Most likely there is a bug in stratis, the program "
                        "you are running.")


def get_errors(exc):
    """
    Generates a sequence of exceptions starting with exc and following the chain
    of causes.
    """
    while True:
        yield exc
        exc = getattr(exc, "__cause__") or getattr(exc, "__context__")
        if exc is None:
            return


def get_error_msgs(errors):
    """
    Generates messages from a sequence of errors. Omits empty messages
    """
    for error in errors:
        if isinstance(error, dbus.exceptions.DBusException):
            error_str = error.get_dbus_message()
        else:
            error_str = str(error)
        if error_str is not None and error_str != "":
            yield error_str


# pylint: disable=too-many-return-statements
def interpret_errors(errors):
    """
    Laboriously add best guesses at the cause of the error, based on
    developer knowledge and possibly further information that is gathered
    in this method.

    :param errors: the chain of errors
    :type errors: list of Exception
    :returns: None if no interpretation found, otherwise str
    """
    # pylint: disable=fixme
    # TODO: This method is extremely rudimentary. It should not be extended
    # using exactly the structure it has now.
    try:
        # Inspect top-most error after StratisCliActionError
        error = errors[1]
        if isinstance(error, AttributeError):
            import traceback
            frame = traceback.extract_tb(error.__traceback__)[-1]
            fmt_str = (
                "Most likely there is an error in the source at line %d "
                "in file %s. The text of the line is \"%s\".")
            return fmt_str % (frame.lineno, frame.filename, frame.line)

        if isinstance(error,
                      DbusClientUniqueResultError) and error.result == []:
            fmt_str = "Most likely you specified a %s which does not exist."
            return fmt_str % interface_name_to_common_name(
                error.interface_name)

        if isinstance(error, DbusClientUnknownSearchPropertiesError):
            return _STRATIS_CLI_BUG_MSG
        if isinstance(error, DbusClientMissingSearchPropertiesError):
            return _DBUS_INTERFACE_MSG
        if isinstance(error, DbusClientMissingInterfaceError):
            return _STRATIS_CLI_BUG_MSG
        if isinstance(error, DbusClientMissingPropertyError):
            return _DBUS_INTERFACE_MSG

        # Inspect lowest error
        error = errors[-1]
        if isinstance(error, dbus.exceptions.DBusException) and \
            error.get_dbus_name() == \
            'org.freedesktop.DBus.Error.AccessDenied':
            return "Most likely stratis has insufficient permissions for the action requested."
        # We have observed two causes of this problem. The first is that
        # stratisd is not running at all. The second is that stratisd has not
        # yet established its D-Bus service.
        if isinstance(error, dbus.exceptions.DBusException) and \
            error.get_dbus_name() == \
            'org.freedesktop.DBus.Error.NameHasNoOwner':
            return "Most likely stratis is unable to connect to the stratisd D-Bus service."

        return None
    # pylint: disable=broad-except
    except Exception:
        return None


def generate_error_message(errors):
    """
    Generate an error message from the given errors.

    :param errors: a list of exceptions
    :type errors: list of Exception

    :returns: str

    Precondition: len(errors) > 0
    """
    # Skip message from first error, which is StratisCliActionError.
    # This error just tells what the command line arguments were and what
    # the resulting parser namespace was, which is probably not interesting
    # to the user.
    error_msgs = [msg for msg in get_error_msgs(errors[1:])]
    if error_msgs == []:
        # It is unlikely that, within the whole chain of errors, there
        # will be no message that is not an empty string. If there is
        # there is some program error, so just raise the exception.
        raise errors[0]

    return ("%s    which in turn caused:%s" % (os.linesep, os.linesep)).join(
        reversed(error_msgs))


def handle_error(err):
    """
    Do the right thing with the given error, which may be the head of an error
    chain.

    :param Exception err: an exception
    """

    errors = [error for error in get_errors(err)]

    error_msg = generate_error_message(errors)

    explanation = interpret_errors(errors)

    exit_msg = "Execution failure caused by:%s%s" % (os.linesep, error_msg) + \
            ("" if explanation is None else "%s%s%s" % (os.linesep, os.linesep, explanation))

    sys.exit(exit_msg)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Miscellaneous methods to support testing.
"""

import os
import string
import subprocess
import sys
import time
import unittest

from hypothesis import strategies

from stratis_cli import run

try:
    _STRATISD = os.environ['STRATISD']
except KeyError:
    message = "STRATISD environment variable must be set to absolute path of stratisd executable"
    sys.exit(message)


def _device_list(minimum):
    """
    Get a device generating strategy.

    :param int minimum: the minimum number of devices, must be at least 0
    """
    return strategies.lists(
        strategies.text(alphabet=string.ascii_letters + "/", min_size=1),
        min_size=minimum)


class _Service():
    """
    Handle starting and stopping the Rust service.
    """

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        self._stratisd = subprocess.Popen([os.path.join(_STRATISD), '--sim'])
        time.sleep(1)

    def tearDown(self):
        """
        Stop the stratisd simulator and daemon.
        """
        # pylint: disable=no-member
        self._stratisd.terminate()
        self._stratisd.wait()


class SimTestCase(unittest.TestCase):
    """
    A SimTestCase must always start and stop stratisd (simulator vesion).
    """

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        self._service = _Service()
        self._service.setUp()

    def tearDown(self):
        """
        Stop the stratisd simulator and daemon.
        """
        self._service.tearDown()


RUNNER = run()



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'stratisd'.
"""

import unittest

import dbus

from stratis_cli._errors import StratisCliActionError

from ._misc import RUNNER
from ._misc import SimTestCase


class StratisTestCase(SimTestCase):
    """
    Test meta information about stratisd.
    """
    _MENU = ['--propagate', 'daemon']

    def testStratisVersion(self):
        """
        Getting version should just succeed.
        """
        command_line = self._MENU + ['version']
        RUNNER(command_line)

    def testStratisRedundancy(self):
        """
        Getting redundancy should just succeed.
        """
        command_line = self._MENU + ['redundancy']
        RUNNER(command_line)

    def testStratisNoOptions(self):
        """
        Exactly one option should be set, this should succeed, but print help.
        """
        command_line = self._MENU
        RUNNER(command_line)

    def testStratisTwoOptions(self):
        """
        Exactly one option should be set, so this should fail,
        but only because redundancy accepts no arguments.
        """
        command_line = self._MENU + ['redundancy', 'version']
        with self.assertRaises(SystemExit):
            RUNNER(command_line)


class PropagateTestCase(unittest.TestCase):
    """
    Verify correct operation of --propagate flag.
    """

    def testPropagate(self):
        """
        If propagate is set, the expected exception will propagate.
        """
        command_line = ['--propagate', 'daemon', 'version']
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, dbus.exceptions.DBusException)

    def testNotPropagate(self):
        """
        If propagate is not set, the exception will be SystemExit.
        """
        command_line = ['daemon', 'version']
        with self.assertRaises(SystemExit):
            RUNNER(command_line)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'snapshot'.
"""

from dbus_client_gen import DbusClientUniqueResultError

from stratis_cli._errors import StratisCliActionError

from .._misc import _device_list
from .._misc import RUNNER
from .._misc import SimTestCase

_DEVICE_STRATEGY = _device_list(1)


class SnapshotTestCase(SimTestCase):
    """
    Test creating a snapshot of a filesystem in a pool.  In this case
    the snapshot should be created and no error raised.
    """
    _MENU = ['--propagate', 'filesystem', 'snapshot']
    _POOLNAME = 'deadpool'
    _SNAPNAME = 'snapfs'
    _FSNAME = 'fs'

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] + \
                _DEVICE_STRATEGY.example()
        RUNNER(command_line)
        command_line = ['filesystem', 'create', self._POOLNAME, self._FSNAME]
        RUNNER(command_line)

    def testSnapshot(self):
        """
        Creation of the snapshot should succeed since origin pool/filesytem is available.
        """
        command_line = self._MENU + [
            self._POOLNAME, self._FSNAME, self._SNAPNAME
        ]
        RUNNER(command_line)


class Snapshot1TestCase(SimTestCase):
    """
    Test creating a snapshot w/out a pool.
    """
    _MENU = ['--propagate', 'filesystem', 'snapshot']
    _POOLNAME = 'nopool'
    _SNAPNAME = 'snapfs'
    _FSNAME = 'fs'

    def testCreation(self):
        """
        Creation of the snapshot must fail since specified pool does not exist.
        """
        command_line = self._MENU + [
            self._POOLNAME, self._FSNAME, self._SNAPNAME
        ]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)


class Snapshot2TestCase(SimTestCase):
    """
    Test creating a snapshot w/out a filesystem.
    """
    _MENU = ['--propagate', 'filesystem', 'snapshot']
    _POOLNAME = 'pool'
    _FSNAME = 'fs'
    _SNAPNAME = 'snapfs'

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] + \
                _DEVICE_STRATEGY.example()
        RUNNER(command_line)

    def testCreation(self):
        """
        Creation of the snapshot must fail since filesystem does not exist.
        """
        command_line = self._MENU + [
            self._POOLNAME, self._FSNAME, self._SNAPNAME
        ]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'rename'.
"""

from dbus_client_gen import DbusClientUniqueResultError

from stratis_cli._errors import StratisCliActionError

from .._misc import RUNNER
from .._misc import SimTestCase
from .._misc import _device_list

_DEVICE_STRATEGY = _device_list(1)


class Rename1TestCase(SimTestCase):
    """
    Test 'rename' when pool is non-existant.
    """
    _MENU = ['--propagate', 'pool', 'rename']
    _POOLNAME = 'deadpool'
    _NEW_POOLNAME = 'livepool'

    def testRename(self):
        """
        This should fail because original name does not exist.
        """
        command_line = self._MENU + [self._POOLNAME, self._NEW_POOLNAME]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)

    def testSameName(self):
        """
        Renaming to itself will fail because the pool does not exist.
        """
        command_line = self._MENU + [self._POOLNAME, self._POOLNAME]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)


class Rename2TestCase(SimTestCase):
    """
    Test 'rename' when pool exists.
    """
    _MENU = ['--propagate', 'pool', 'rename']
    _POOLNAME = 'deadpool'
    _NEW_POOLNAME = 'livepool'

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        RUNNER(command_line)

    def testRename(self):
        """
        This should succeed because pool exists.
        """
        command_line = self._MENU + [self._POOLNAME, self._NEW_POOLNAME]
        RUNNER(command_line)

    def testSameName(self):
        """
        This should succeed, because renaming to self makes sense.
        """
        command_line = self._MENU + [self._POOLNAME, self._POOLNAME]
        RUNNER(command_line)

    def testNonExistentPool(self):
        """
        This should fail, because this pool is not there.
        """
        command_line = self._MENU + ["nopool", self._POOLNAME]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Definition of pool actions to display in the CLI.
"""

from .._actions import TopActions

POOL_SUBCMDS = [
    ('create',
     dict(
         help='Create a pool',
         args=[
             ('pool_name', dict(
                 action='store',
                 help='Name of new pool',
             )),
             ('blockdevs',
              dict(
                  help='Create the pool using these block devs',
                  nargs='+',
              )),
             ('--redundancy',
              dict(
                  action='store',
                  choices=['none'],
                  default='none',
                  help="Redundancy level for this pool",
              )),
         ],
         func=TopActions.create_pool,
     )),
    ('list',
     dict(
         help="List pools",
         description="Lists Stratis pools that exist on the system",
         func=TopActions.list_pools,
     )),
    ('destroy',
     dict(
         help='Destroy a pool',
         args=[
             ('pool_name', dict(
                 action='store',
                 help='pool name',
             )),
         ],
         func=TopActions.destroy_pool,
     )),
    ('rename',
     dict(
         help='Rename a pool',
         args=[
             ('current', dict(
                 action='store',
                 help='Current pool name',
             )),
             ('new', dict(
                 action='store',
                 help='New pool name',
             )),
         ],
         func=TopActions.rename_pool,
     )),
    ('add-data',
     dict(
         help=
         "Add one or more blockdevs to an existing pool for use as data storage",
         args=[
             ('pool_name', dict(
                 action='store',
                 help='Pool name',
             )),
             ('device',
              dict(
                  help='Block devices to add to the pool',
                  metavar='blockdev',
                  nargs='+')),
         ],
         func=TopActions.add_data_device)),
    ('add-cache',
     dict(
         help="Add one or more blockdevs to an existing pool for use as cache",
         args=[
             ('pool_name', dict(
                 action='store',
                 help='Pool name',
             )),
             ('device',
              dict(
                  help='Block devices to add to the pool as cache',
                  metavar='blockdev',
                  nargs='+')),
         ],
         func=TopActions.add_cache_device)),
]



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'list'.
"""

from .._misc import _device_list
from .._misc import RUNNER
from .._misc import SimTestCase

_DEVICE_STRATEGY = _device_list(1)


class ListTestCase(SimTestCase):
    """
    Test 'list'.
    """
    _MENU = ['--propagate', 'pool', 'list']

    def testList(self):
        """
        List should just succeed, even though there is nothing to list.
        """
        command_line = self._MENU
        RUNNER(command_line)

    def testListDefault(self):
        """
        Test default listing action when "list" is not specified.

        List should just succeed, even though there is nothing to list.
        """
        command_line = self._MENU[:-1]
        RUNNER(command_line)


class List2TestCase(SimTestCase):
    """
    Test 'list' with something actually to list.
    """
    _MENU = ['--propagate', 'pool', 'list']
    _POOLNAME = 'deadpool'

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        RUNNER(command_line)

    def testList(self):
        """
        List should just succeed.
        """
        command_line = self._MENU
        RUNNER(command_line)

    def testListDefault(self):
        """
        Test default listing action when "list" is not specified.
        """
        command_line = self._MENU[:-1]
        RUNNER(command_line)



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
XML interface specifications.
"""

import xml.etree.ElementTree as ET

from dbus_client_gen import managed_object_class
from dbus_client_gen import mo_query_builder
from dbus_client_gen import DbusClientGenerationError

from dbus_python_client_gen import make_class
from dbus_python_client_gen import DPClientGenerationError

from .._errors import StratisCliGenerationError
from .._errors import StratisCliValueError

from ._constants import DBUS_TIMEOUT_SECONDS

SPECS = {
    "org.freedesktop.DBus.ObjectManager":
    """
<interface name="org.freedesktop.DBus.ObjectManager">
<method name="GetManagedObjects">
<arg name="objpath_interfaces_and_properties" type="a{oa{sa{sv}}}" direction="out"/>
</method>
</interface>
""",
    "org.storage.stratis1.Manager":
    """
<interface name="org.storage.stratis1.Manager">
<method name="ConfigureSimulator">
<arg name="denominator" type="u" direction="in"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="CreatePool">
<arg name="name" type="s" direction="in"/>
<arg name="redundancy" type="(bq)" direction="in"/>
<arg name="devices" type="as" direction="in"/>
<arg name="result" type="(oao)" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="DestroyPool">
<arg name="pool" type="o" direction="in"/>
<arg name="action" type="b" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<property name="Version" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
</interface>
""",
    "org.storage.stratis1.pool":
    """
<interface name="org.storage.stratis1.pool">
<method name="AddCacheDevs">
<arg name="devices" type="as" direction="in"/>
<arg name="results" type="ao" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="AddDataDevs">
<arg name="devices" type="as" direction="in"/>
<arg name="results" type="ao" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="CreateFilesystems">
<arg name="specs" type="as" direction="in"/>
<arg name="filesystems" type="a(os)" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="DestroyFilesystems">
<arg name="filesystems" type="ao" direction="in"/>
<arg name="results" type="as" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="SetName">
<arg name="name" type="s" direction="in"/>
<arg name="action" type="b" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<method name="SnapshotFilesystem">
<arg name="origin" type="o" direction="in"/>
<arg name="snapshot_name" type="s" direction="in"/>
<arg name="result" type="o" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<property name="Name" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
</property>
<property name="TotalPhysicalSize" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false"/>
</property>
<property name="TotalPhysicalUsed" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false"/>
</property>
<property name="Uuid" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="State" type="q" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
</property>
<property name="ExtendState" type="q" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
</property>
<property name="SpaceState" type="q" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
</property>
</interface>
""",
    "org.storage.stratis1.filesystem":
    """
<interface name="org.storage.stratis1.filesystem">
<method name="SetName">
<arg name="name" type="s" direction="in"/>
<arg name="action" type="b" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<property name="Created" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="Devnode" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="Name" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
</property>
<property name="Pool" type="o" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="Used" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false"/>
</property>
<property name="Uuid" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
</interface>
""",
    "org.storage.stratis1.blockdev":
    """
<interface name="org.storage.stratis1.blockdev">
<method name="SetUserInfo">
<arg name="id" type="s" direction="in"/>
<arg name="changed" type="b" direction="out"/>
<arg name="return_code" type="q" direction="out"/>
<arg name="return_string" type="s" direction="out"/>
</method>
<property name="Devnode" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="HardwareInfo" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="InitializationTime" type="t" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="Pool" type="o" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
<property name="State" type="q" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
</property>
<property name="Tier" type="q" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false"/>
</property>
<property name="TotalPhysicalSize" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false"/>
</property>
<property name="UserInfo" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="false"/>
</property>
<property name="Uuid" type="s" access="read">
<annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const"/>
</property>
</interface>
""",
}

_FILESYSTEM_INTERFACE = 'org.storage.stratis1.filesystem'
_POOL_INTERFACE = 'org.storage.stratis1.pool'
_BLOCKDEV_INTERFACE = 'org.storage.stratis1.blockdev'


def interface_name_to_common_name(interface_name):
    """
    Maps a D-Bus interface name to the common name that identifies the type
    of stratisd thing that the interface represents.

    :param str interface_name: the interface name
    :returns: a common name
    :rtype: str
    """
    if interface_name == _BLOCKDEV_INTERFACE:
        return "block device"

    if interface_name == _FILESYSTEM_INTERFACE:
        return "filesystem"

    if interface_name == _POOL_INTERFACE:
        return "pool"

    raise StratisCliValueError(interface_name, "interface_name")


try:
    filesystem_spec = ET.fromstring(SPECS[_FILESYSTEM_INTERFACE])
    Filesystem = make_class("Filesystem", filesystem_spec,
                            DBUS_TIMEOUT_SECONDS)
    MOFilesystem = managed_object_class("MOFilesystem", filesystem_spec)
    filesystems = mo_query_builder(filesystem_spec)

    pool_spec = ET.fromstring(SPECS[_POOL_INTERFACE])
    Pool = make_class("Pool", pool_spec, DBUS_TIMEOUT_SECONDS)
    MOPool = managed_object_class("MOPool", pool_spec)
    pools = mo_query_builder(pool_spec)

    blockdev_spec = ET.fromstring(SPECS[_BLOCKDEV_INTERFACE])
    MODev = managed_object_class("MODev", blockdev_spec)
    devs = mo_query_builder(blockdev_spec)

    Manager = make_class("Manager",
                         ET.fromstring(SPECS['org.storage.stratis1.Manager']),
                         DBUS_TIMEOUT_SECONDS)

    ObjectManager = make_class(
        "ObjectManager",
        ET.fromstring(SPECS['org.freedesktop.DBus.ObjectManager']),
        DBUS_TIMEOUT_SECONDS)

except DPClientGenerationError as err:
    raise StratisCliGenerationError(
        "Failed to generate some class needed for invoking dbus-python methods"
    ) from err
except DbusClientGenerationError as err:
    raise StratisCliGenerationError(
        "Failed to generate some class needed for examining D-Bus data"
    ) from err



#!/usr/bin/python

import argparse
import subprocess
import sys

arg_map = {
    "src/stratis_cli": [
        "--reports=no", "--disable=I", "--disable=duplicate-code",
        "--disable=invalid-name",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'"
    ],
    "tests": [
        "--reports=no", "--disable=I", "--disable=duplicate-code",
        "--disable=invalid-name",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'"
    ],
    "bin/stratis": [
        "--reports=no",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'"
    ],
}


def get_parser():
    """
    Generate an appropriate parser.

    :returns: an argument parser
    :rtype: `ArgumentParser`
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "package",
        choices=arg_map.keys(),
        help="designates the package to test")
    parser.add_argument("--ignore", help="ignore these files")
    return parser


def get_command(namespace):
    """
    Get the pylint command for these arguments.

    :param `Namespace` namespace: the namespace
    """
    cmd = ["pylint", namespace.package] + arg_map[namespace.package]
    if namespace.ignore:
        cmd.append("--ignore=%s" % namespace.ignore)
    return cmd


def main():
    args = get_parser().parse_args()
    return subprocess.call(get_command(args), stdout=sys.stdout)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Miscellaneous top-level actions.
"""

from justbytes import Range

from .._errors import StratisCliEngineError

from .._stratisd_constants import StratisdErrors

from ._connection import get_object
from ._constants import TOP_OBJECT
from ._constants import SECTOR_SIZE
from ._data import Manager
from ._data import MOPool
from ._data import ObjectManager
from ._data import Pool
from ._data import pools
from ._formatting import print_table


class TopActions():
    """
    Top level actions.
    """

    @staticmethod
    def create_pool(namespace):
        """
        Create a stratis pool.

        :raises StratisCliEngineError:
        """
        proxy = get_object(TOP_OBJECT)

        (_, rc, message) = Manager.Methods.CreatePool(
            proxy, {
                'name': namespace.pool_name,
                'redundancy': (True, 0),
                'devices': namespace.blockdevs
            })

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def list_pools(_):
        """
        List all stratis pools.

        :raises StratisCliEngineError:
        """
        proxy = get_object(TOP_OBJECT)

        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        mopools = (MOPool(info) for _, info in pools().search(managed_objects))
        tables = [[
            mopool.Name(),
            str(Range(mopool.TotalPhysicalSize(), SECTOR_SIZE)),
            str(Range(mopool.TotalPhysicalUsed(), SECTOR_SIZE)),
        ] for mopool in mopools]
        print_table(['Name', 'Total Physical Size', 'Total Physical Used'],
                    sorted(tables, key=lambda entry: entry[0]),
                    ['<', '>', '>'])

    @staticmethod
    def destroy_pool(namespace):
        """
        Destroy a stratis pool.

        If no pool exists, the method succeeds.

        :raises StratisCliEngineError:
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = \
           Manager.Methods.DestroyPool(proxy, {'pool': pool_object_path})

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def rename_pool(namespace):
        """
        Rename a pool.
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.current
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = Pool.Methods.SetName(
            get_object(pool_object_path), {'name': namespace.new})

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def add_data_device(namespace):
        """
        Add a device to a pool.
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = Pool.Methods.AddDataDevs(
            get_object(pool_object_path), {'devices': namespace.device})
        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def add_cache_device(namespace):
        """
        Add a device to a pool.
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = Pool.Methods.AddCacheDevs(
            get_object(pool_object_path), {'devices': namespace.device})
        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'destroy'.
"""

from dbus_client_gen import DbusClientUniqueResultError

from stratis_cli._errors import StratisCliActionError
from stratis_cli._errors import StratisCliEngineError

from stratis_cli._stratisd_constants import StratisdErrors

from .._misc import _device_list
from .._misc import RUNNER
from .._misc import SimTestCase

_DEVICE_STRATEGY = _device_list(1)


class Destroy1TestCase(SimTestCase):
    """
    Test 'destroy' on empty database.

    'destroy' should always fail if pool is missing.
    """
    _MENU = ['--propagate', 'pool', 'destroy']
    _POOLNAME = 'deadpool'

    def testExecution(self):
        """
        Destroy should fail because there is no object path for the pool.
        """
        command_line = self._MENU + [self._POOLNAME]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)


class Destroy2TestCase(SimTestCase):
    """
    Test 'destroy' on database which contains the given pool.
    """
    _MENU = ['--propagate', 'pool', 'destroy']
    _POOLNAME = 'deadpool'

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        RUNNER(command_line)

    def testExecution(self):
        """
        The pool was just created, so must be destroyable.
        """
        command_line = self._MENU + [self._POOLNAME]
        RUNNER(command_line)


class Destroy3TestCase(SimTestCase):
    """
    Test 'destroy' on database which contains the given pool with a volume.
    """
    _MENU = ['--propagate', 'pool', 'destroy']
    _POOLNAME = 'deadpool'
    _VOLNAME = 'vol'

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        RUNNER(command_line)

        command_line = ['filesystem', 'create', self._POOLNAME, self._VOLNAME]
        RUNNER(command_line)

    def testExecution(self):
        """
        This should fail since it has a filesystem.
        """
        command_line = self._MENU + [self._POOLNAME]
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, StratisCliEngineError)
        self.assertEqual(cause.rc, StratisdErrors.BUSY)

    def testWithFilesystemRemoved(self):
        """
        This should succeed since the filesystem is removed first.
        """
        command_line = ['filesystem', 'destroy', self._POOLNAME, self._VOLNAME]
        RUNNER(command_line)
        command_line = self._MENU + [self._POOLNAME]
        RUNNER(command_line)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Formatting for tables.
"""

import sys


def print_table(column_headings, row_entries, alignment, file=sys.stdout):
    """
    Given the column headings and the row_entries, print a table.
    Align according to alignment specification and always pad with 2 spaces.

    :param column_headings: the column headings
    :type column_headings: list of str
    :param row_entries: a list of the row entries
    :type row_entries: list of list of str
    :param alignment: the alignment indicator for each key, '<', '>', '^', '='
    :type alignment: list of str

    Precondition: len(column_headings) == len(alignment) == len of each entry
    in row_entries.
    """

    num_columns = len(column_headings)

    column_lengths = [
        max(
            max((len(e[index]) for e in row_entries), default=0),
            len(column_headings[index])) + 2 for index in range(num_columns)
    ]

    for index in range(num_columns):
        line = '{0:{align}{width}}'.format(
            column_headings[index],
            align=alignment[index],
            width=column_lengths[index])
        print(line, end='', file=file)
    print(file=file)

    for row in row_entries:
        for index in range(num_columns):
            line = '{0:{align}{width}}'.format(
                row[index],
                align=alignment[index],
                width=column_lengths[index])
            print(line, end='', file=file)
        print(file=file)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Shared utilities.
"""

from ._data import MOPool
from ._data import pools


def get_objects(namespace, pool_name_key, managed_objects, search_function,
                constructor):
    """
    Get objects specified by namespace and designated pool_name_key.
    If no pool name key is specified in the namespace, get objects for all
    pools. Returns objects of interest and also a map from pool object path
    to name for all the pools which have objects in the returned list of
    objects.

    :param Namesapce namespace: the parser namespace
    :param str pool_name_key: the key in the namespace for the pool name
    :param dict managed_objects: the result of a GetManagedObjects call
    :param search_function: function that finds objects of a given type
    :type search_function: dict -> (dict -> list of str * dict)
    :param constructor: function to wrap dict info about a given object

    :returns: objects of interest and a map from pool object path to name
    :rtype: list * dict
    """
    pool_name = getattr(namespace, pool_name_key, None)
    if pool_name is not None:
        (parent_pool_object_path, _) = next(
            pools(props={
                'Name': pool_name
            }).require_unique_match(True).search(managed_objects))

        properties = {"Pool": parent_pool_object_path}
        path_to_name = {parent_pool_object_path: namespace.pool_name}
    else:
        properties = {}
        path_to_name = dict((path, MOPool(info).Name())
                            for path, info in pools().search(managed_objects))

    objects = [
        constructor(info) for _, info in search_function(props=properties, )
        .search(managed_objects)
    ]

    return (objects, path_to_name)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'destroy'.
"""

import unittest

from dbus_client_gen import DbusClientUniqueResultError

from stratis_cli._errors import StratisCliActionError

from .._misc import _device_list
from .._misc import RUNNER
from .._misc import SimTestCase

_DEVICE_STRATEGY = _device_list(1)


@unittest.skip(
    "Temporarily unable to create multiple filesystems at same time")
class DestroyTestCase(SimTestCase):
    """
    Test destroying a volume when the pool does not exist. In this case,
    an error should be raised, as the request is non-sensical.
    """
    _MENU = ['--propagate', 'filesystem', 'destroy']
    _POOLNAME = 'deadpool'
    _VOLNAMES = ['oubliette', 'mnemosyne']

    def testDestroy(self):
        """
        Destruction of the volume must fail since pool is not specified.
        """
        command_line = self._MENU + [self._POOLNAME] + self._VOLNAMES
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)


@unittest.skip(
    "Temporarily unable to create multiple filesystems at same time")
class Destroy2TestCase(SimTestCase):
    """
    Test destroying a volume when the pool does exist but the volume does not.
    In this case, no error should be raised.
    """
    _MENU = ['--propagate', 'filesystem', 'destroy']
    _POOLNAME = 'deadpool'
    _VOLNAMES = ['oubliette', 'mnemosyne']

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        RUNNER(command_line)

    def testDestroy(self):
        """
        Destruction of the volume must succeed since pool exists and at end
        volume is gone.
        """
        command_line = self._MENU + [self._POOLNAME] + self._VOLNAMES
        RUNNER(command_line)


@unittest.skip(
    "Temporarily unable to create multiple filesystems at same time")
class Destroy3TestCase(SimTestCase):
    """
    Test destroying a volume when the pool does exist and the volume does as
    well. In this case, the volumes should all be destroyed, and no error
    raised as there is no data on the volumes.
    """
    _MENU = ['--propagate', 'filesystem', 'destroy']
    _POOLNAME = 'deadpool'
    _VOLNAMES = ['oubliette', 'mnemosyne']

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        command_line = ['pool', 'create', self._POOLNAME] + \
                _DEVICE_STRATEGY.example()
        RUNNER(command_line)

        command_line = ['filesystem', 'create', self._POOLNAME] + \
                self._VOLNAMES
        RUNNER(command_line)

    def testDestroy(self):
        """
        Destruction of the volume must succeed since pool exists and at end
        volume is gone.
        """
        command_line = self._MENU + [self._POOLNAME] + self._VOLNAMES
        RUNNER(command_line)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Miscellaneous logical actions.
"""

from justbytes import Range
from dateutil import parser as date_parser

from .._errors import StratisCliEngineError

from .._stratisd_constants import StratisdErrors

from ._connection import get_object
from ._constants import TOP_OBJECT
from ._data import MOFilesystem
from ._data import ObjectManager
from ._data import Pool
from ._data import Filesystem
from ._data import filesystems
from ._data import pools
from ._formatting import print_table
from ._util import get_objects


class LogicalActions():
    """
    Actions on the logical aspects of a pool.
    """

    @staticmethod
    def create_volumes(namespace):
        """
        Create volumes in a pool.

        :raises StratisCliEngineError:
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = Pool.Methods.CreateFilesystems(
            get_object(pool_object_path), {'specs': namespace.fs_name})

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def list_volumes(namespace):
        """
        List the volumes in a pool.
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (mofilesystems, path_to_name) = get_objects(
            namespace, "pool_name", managed_objects, filesystems, MOFilesystem)

        tables = [[
            path_to_name[mofilesystem.Pool()],
            mofilesystem.Name(),
            str(Range(mofilesystem.Used())),
            date_parser.parse(mofilesystem.Created()).astimezone().strftime(
                "%b %d %Y %H:%M"),
            mofilesystem.Devnode(),
            mofilesystem.Uuid(),
        ] for mofilesystem in mofilesystems]

        print_table(['Pool Name', 'Name', 'Used', 'Created', 'Device', 'UUID'],
                    sorted(tables, key=lambda entry: entry[0]),
                    ['<', '<', '<', '<', '<', '<'])

    @staticmethod
    def destroy_volumes(namespace):
        """
        Destroy volumes in a pool.

        :raises StratisCliEngineError:
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))
        fs_object_paths = [
            op for name in namespace.fs_name
            for (op, _) in filesystems(props={
                'Name': name,
                'Pool': pool_object_path
            }).search(managed_objects)
        ]

        (_, rc, message) = Pool.Methods.DestroyFilesystems(
            get_object(pool_object_path), {'filesystems': fs_object_paths})

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def snapshot_filesystem(namespace):
        """
        Snapshot filesystem in a pool.

        :raises StratisCliEngineError:
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))
        (origin_fs_object_path, _) = next(
            filesystems(props={
                'Name': namespace.origin_name,
                'Pool': pool_object_path
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = Pool.Methods.SnapshotFilesystem(
            get_object(pool_object_path), {
                'origin': origin_fs_object_path,
                'snapshot_name': namespace.snapshot_name
            })

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)

    @staticmethod
    def rename_fs(namespace):
        """
        Rename a filesystem.
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        (pool_object_path, _) = next(
            pools(props={
                'Name': namespace.pool_name
            }).require_unique_match(True).search(managed_objects))
        (fs_object_path, _) = next(
            filesystems(props={
                'Name': namespace.fs_name,
                'Pool': pool_object_path
            }).require_unique_match(True).search(managed_objects))

        (_, rc, message) = Filesystem.Methods.SetName(
            get_object(fs_object_path), {'name': namespace.new_name})

        if rc != StratisdErrors.OK:
            raise StratisCliEngineError(rc, message)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'create'.
"""

from dbus_client_gen import DbusClientUniqueResultError

from stratis_cli._errors import StratisCliActionError

from .._misc import _device_list
from .._misc import RUNNER
from .._misc import SimTestCase

_DEVICE_STRATEGY = _device_list(1)


class AddDataTestCase(SimTestCase):
    """
    Test adding devices to a non-existant pool.
    """
    _MENU = ['--propagate', 'pool', 'add-data']
    _POOLNAME = 'deadpool'

    def testAdd(self):
        """
        Adding the devices must fail since the pool does not exist.
        """
        command_line = self._MENU + [self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)


class AddCacheTestCase(SimTestCase):
    """
    Test adding devices to a non-existant pool.
    """
    _MENU = ['--propagate', 'pool', 'add-cache']
    _POOLNAME = 'deadpool'

    def testAdd(self):
        """
        Adding the devices must fail since the pool does not exist.
        """
        command_line = self._MENU + [self._POOLNAME] \
            + _DEVICE_STRATEGY.example()
        with self.assertRaises(StratisCliActionError) as context:
            RUNNER(command_line)
        cause = context.exception.__cause__
        self.assertIsInstance(cause, DbusClientUniqueResultError)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Error heirarchy for stratis cli.
"""

from ._stratisd_constants import STRATISD_ERROR_TO_NAME


class StratisCliError(Exception):
    """
    Top-level stratis cli error.
    """


class StratisCliRuntimeError(StratisCliError):
    """
    Exception raised during runtime.
    """


class StratisCliValueError(StratisCliRuntimeError):
    """
    Error raised when a parameter has an unacceptable value.
    """

    def __init__(self, value, param):
        """ Initializer.

            :param object value: the value
            :param str param: the parameter
        """
        # pylint: disable=super-init-not-called
        self._value = value
        self._param = param

    def __str__(self):
        return "value '%s' for parameter %s is unacceptable" % (self._value,
                                                                self._param)


class StratisCliEngineError(StratisCliRuntimeError):
    """
    Raised if there was a failure due to an error in stratisd's engine.
    """

    def __init__(self, rc, message):
        """ Initializer.

            :param rc int: the error code returned by the engine
            :param str message: whatever message accompanied the error code
        """
        # pylint: disable=super-init-not-called
        self.rc = rc
        self.message = message

    def __str__(self):
        return "%s: %s" % (STRATISD_ERROR_TO_NAME(self.rc), self.message)


class StratisCliActionError(StratisCliRuntimeError):
    """
    Raised if an action selected by the parser failed.
    """

    def __init__(self, command_line_args, namespace):
        """
        Initialize with parser-returned namespace.

        :param command_line_args: the arguments passed on the command line
        :type command_line_args: list of str
        :param Namespace namespace: the namespace constructed by the parser
        """
        # pylint: disable=super-init-not-called
        self.command_line_args = command_line_args
        self.namespace = namespace

    def __str__(self):
        fmt_str = ("Action selected by command-line arguments %s which were "
                   "parsed to %s failed")
        return fmt_str % (self.command_line_args, self.namespace)


class StratisCliGenerationError(StratisCliError):
    """
    Exception that occurs during generation of classes.
    """



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Miscellaneous physical actions.
"""

from justbytes import Range

from ._connection import get_object
from ._constants import SECTOR_SIZE
from ._constants import TOP_OBJECT
from ._constants import UNKNOWN_VALUE_MARKER
from ._data import devs
from ._data import MODev
from ._data import ObjectManager
from ._formatting import print_table
from ._util import get_objects


def state_val_to_string(val):
    """
    Convert a blockdev state enumerated value to a string.
    """
    states = ["Missing", "Bad", "Spare", "Not-in-use", "In-use"]
    try:
        return states[val]
    except IndexError:
        return UNKNOWN_VALUE_MARKER


def tier_val_to_string(val):
    """
    Convert a blockdev tier enumerated value to a string.
    """
    states = ["Data", "Cache"]
    try:
        return states[val]
    except IndexError:
        return UNKNOWN_VALUE_MARKER


class PhysicalActions():
    """
    Actions on the physical aspects of a pool.
    """
    # pylint: disable=too-few-public-methods

    @staticmethod
    def list_pool(namespace):
        """
        List devices. If a pool is specified in the namespace, list devices
        for that pool. Otherwise, list all devices for all pools.
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        (modevs, path_to_name) = get_objects(namespace, "pool_name",
                                             managed_objects, devs, MODev)

        tables = [[
            path_to_name[modev.Pool()],
            modev.Devnode(),
            str(Range(modev.TotalPhysicalSize(), SECTOR_SIZE)),
            state_val_to_string(modev.State()),
            tier_val_to_string(modev.Tier())
        ] for modev in modevs]
        print_table(
            ["Pool Name", "Device Node", "Physical Size", "State", "Tier"],
            sorted(tables, key=lambda entry: (entry[0], entry[1])),
            ['<', '<', '>', '>', '>'])



import os
import sys
import setuptools
if sys.version_info[0] < 3:
    from codecs import open


def local_file(name):
    return os.path.relpath(os.path.join(os.path.dirname(__file__), name))


README = local_file("README.rst")

with open(local_file("src/stratis_cli/_version.py")) as o:
    exec(o.read())

setuptools.setup(
    name='stratis-cli',
    version=__version__,
    author='Anne Mulhern',
    author_email='amulhern@redhat.com',
    description='prototype stratis cli',
    long_description=open(README, encoding='utf-8').read(),
    platforms=['Linux'],
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha', 'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux', 'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Filesystems', 'Topic :: Systems Administration'
    ],
    install_requires=[
        'dbus-client-gen>=0.4',
        'dbus-python-client-gen>=0.5',
        'justbytes==0.11',
        'python-dateutil',
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    scripts=['bin/stratis'])



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Definition of filesystem actions to display in the CLI.
"""

from .._actions import LogicalActions

LOGICAL_SUBCMDS = [
    ('create',
     dict(
         help='Create filesystems in a pool',
         args=[
             ('pool_name', dict(
                 action='store',
                 help='pool name',
             )),
             ('fs_name',
              dict(
                  help='Create filesystems in this pool using the given names',
                  nargs='+',
              )),
         ],
         func=LogicalActions.create_volumes)),
    ('snapshot',
     dict(
         help='Snapshot the named filesystem in a pool',
         args=[
             ('pool_name', dict(
                 action='store',
                 help='pool name',
             )),
             ('origin_name', dict(
                 action='store',
                 help='origin name',
             )),
             ('snapshot_name', dict(
                 action='store',
                 help='snapshot name',
             )),
         ],
         func=LogicalActions.snapshot_filesystem)),
    ('list',
     dict(
         help="List filesystems",
         args=[
             ('pool_name',
              dict(action='store', default=None, nargs="?", help='Pool name')),
         ],
         func=LogicalActions.list_volumes)),
    ('destroy',
     dict(
         help='Destroy filesystems in a pool',
         args=[
             ('pool_name', dict(
                 action='store',
                 help='pool name',
             )),
             ('fs_name',
              dict(
                  help='Destroy the named filesystems in this pool',
                  nargs='+',
              )),
         ],
         func=LogicalActions.destroy_volumes)),
    ('rename',
     dict(
         help='Rename a filesystem',
         args=[
             ('pool_name',
              dict(
                  action='store',
                  help='Name of the pool the filesystem is part of',
              )),
             ('fs_name',
              dict(
                  action='store',
                  help='Name of the filesystem to change',
              )),
             ('new_name',
              dict(
                  action='store',
                  help='New name to give that filesystem',
              )),
         ],
         func=LogicalActions.rename_fs,
     )),
]



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Top level parser for Stratis CLI.
"""

import argparse

from .._actions import LogicalActions
from .._actions import PhysicalActions
from .._actions import StratisActions
from .._actions import TopActions

from .._version import __version__

from ._logical import LOGICAL_SUBCMDS
from ._physical import PHYSICAL_SUBCMDS
from ._pool import POOL_SUBCMDS

# pylint: disable=undefined-variable
PRINT_HELP = lambda parser: lambda _: parser.print_help()


def add_args(parser, args=None):
    """
    Call subcommand.add_argument() based on args list.
    """
    if args is not None:
        for name, arg in args:
            parser.add_argument(name, **arg)


def add_subcommand(subparser, cmd):
    """
    Add subcommand to a parser based on a subcommand dict.
    """
    name, info = cmd
    parser = subparser.add_parser(
        name, help=info['help'], aliases=info.get('aliases', []))

    subcmds = info.get('subcmds')
    if subcmds is not None:
        subparsers = parser.add_subparsers(title='subcommands')
        for subcmd in subcmds:
            add_subcommand(subparsers, subcmd)

    add_args(parser, info.get('args', []))

    parser.set_defaults(func=info.get('func', PRINT_HELP(parser)))


DAEMON_SUBCMDS = [
    ('redundancy',
     dict(
         help="Redundancy designations understood by stratisd daemon",
         func=StratisActions.list_stratisd_redundancy,
     )),
    ('version',
     dict(
         help="version of stratisd daemon",
         func=StratisActions.list_stratisd_version,
     )),
]

ROOT_SUBCOMMANDS = [
    ('pool',
     dict(
         help="Perform General Pool Actions",
         subcmds=POOL_SUBCMDS,
         func=TopActions.list_pools,
     )),
    ('blockdev',
     dict(
         help="Commands related to block devices that make up the pool",
         subcmds=PHYSICAL_SUBCMDS,
         func=PhysicalActions.list_pool,
     )),
    ('filesystem',
     dict(
         aliases=['fs'],
         help="Commands related to filesystems allocated from a pool",
         subcmds=LOGICAL_SUBCMDS,
         func=LogicalActions.list_volumes,
     )),
    ('daemon', dict(
        help="Stratis daemon information",
        subcmds=DAEMON_SUBCMDS,
    )),
]

GEN_ARGS = [
    ('--propagate',
     dict(
         action='store_true',
         help='Allow exceptions to propagate',
     )),
]


def gen_parser():
    """
    Make the parser.

    :returns: a fully constructed parser for command-line arguments
    :rtype: ArgumentParser
    """
    parser = argparse.ArgumentParser(
        description="Stratis Storage Manager", prog='stratis')

    # version is special, it has explicit support in argparse
    parser.add_argument('--version', action='version', version=__version__)

    add_args(parser, GEN_ARGS)

    subparsers = parser.add_subparsers(title='subcommands')

    for subcmd in ROOT_SUBCOMMANDS:
        add_subcommand(subparsers, subcmd)

    parser.set_defaults(func=PRINT_HELP(parser))

    return parser



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Stratisd error classes.
"""

from enum import IntEnum


def value_to_name(klass):
    """
    Generate a function to convert an IntEnum value to its name.

    :param type klass: the class defining the IntEnum
    :returns: a function to convert a single number to a name
    :rtype: int -> str
    """

    def the_func(num):
        """
        Convert the enum value to a string which is just its name.
        Replace underscores in the name with spaces.

        If there is no name for the value, return a special string.

        :param int num: the number to convert
        :returns: the name for the number or an error string
        :rtype: str
        """
        try:
            the_str = str(klass(num)).rsplit('.')[-1].replace("_", " ")
        except ValueError:
            the_str = "Unknown value (%s) for %s constant" % (num,
                                                              klass.__name__)
        return the_str

    return the_func


class StratisdErrors(IntEnum):
    """
    Stratisd Errors
    """
    OK = 0
    ERROR = 1

    ALREADY_EXISTS = 2
    BUSY = 3
    INTERNAL_ERROR = 4
    NOT_FOUND = 5


STRATISD_ERROR_TO_NAME = value_to_name(StratisdErrors)


class RedundancyCodes(IntEnum):
    """
    Redundancy Codes
    """
    NONE = 0


REDUNDANCY_CODE_TO_NAME = value_to_name(RedundancyCodes)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Definition of block device actions to display in the CLI.
"""

from .._actions import PhysicalActions

PHYSICAL_SUBCMDS = [
    ('list',
     dict(
         help="List information about blockdevs in the pool",
         args=[
             ('pool_name',
              dict(action='store', default=None, nargs="?", help='Pool name')),
         ],
         func=PhysicalActions.list_pool)),
]
